with Ada.Calendar;
with Ada.Containers.Indefinite_Ordered_Maps;
with Ada.Directories;
with Ada.Streams.Stream_IO;
with Ada.Strings.Unbounded;
with Ada.Text_IO;
with Scream.Markup;
with Scream.Pages;
with Scream.Parse;
procedure Scream.Build(
	Format : not null access Output.Format'Class;
	Base_Directory : String;
	Force : Boolean;
	Verbose : Boolean;
	Error : out Boolean)
is
	use type Ada.Calendar.Time;
	use type Markup.Page_Access;
	use type Pages.Page_Info_Access;
	
	function To_String (X : Integer) return String is
		Image : constant String := Integer'Image (X);
	begin
		if Image (Image'First) = ' ' then
			return Image (Image'First + 1 .. Image'Last);
		else
			return Image;
		end if;
	end To_String;
	
	package Page_Info_Maps is
		new Ada.Containers.Indefinite_Ordered_Maps (String, Pages.Page_Info_Access);
	package Page_Maps is
		new Ada.Containers.Indefinite_Ordered_Maps (String, Markup.Page_Access);
	
	Page_Info : Page_Info_Maps.Map;
	Modified_Pages : Pages.Page_Sets.Set;
	Build_Pages : Pages.Page_Sets.Set;
	Page_Contents : Page_Maps.Map;
begin
	Error := False;
	-- search directories
	declare
		procedure Search_Pages (Path : in String) is
			Search : Ada.Directories.Search_Type;
			The_Entry : Ada.Directories.Directory_Entry_Type;
		begin
			Ada.Directories.Start_Search (
				Search,
				Directory => Path,
				Pattern => "*." & Pages.Source_Extension,
				Filter => (Ada.Directories.Ordinary_File => True, others => False));
			while Ada.Directories.More_Entries (Search) loop
				Ada.Directories.Get_Next_Entry (Search, The_Entry);
				declare
					Name : constant String := Ada.Directories.Simple_Name (The_Entry);
					Page_Path : constant String :=
						Pages.Compose (
							Containing_Directory => Path,
							Name => Name (Name'First .. Name'Last - (Pages.Source_Extension'Length + 1)));
				begin
					Page_Info_Maps.Insert (
						Page_Info,
						Page_Path,
						new Pages.Page_Info'(Pages.Get_Page_Info (Page_Path)));
				end;
			end loop;
			Ada.Directories.End_Search (Search);
			Ada.Directories.Start_Search (
				Search,
				Directory => Path,
				Pattern => "*",
				Filter => (Ada.Directories.Directory => True, others => False));
			while Ada.Directories.More_Entries (Search) loop
				Ada.Directories.Get_Next_Entry (Search, The_Entry);
				Search_Pages (
					Ada.Directories.Compose (Path, Ada.Directories.Simple_Name (The_Entry)));
			end loop;
			Ada.Directories.End_Search (Search);
		end Search_Pages;
	begin
		Search_Pages (Base_Directory);
	end;
	if Verbose then
		Ada.Text_IO.Put_Line ("all pages:");
		declare
			I : Page_Info_Maps.Cursor := Page_Info.First;
		begin
			while Page_Info_Maps.Has_Element (I) loop
				Ada.Text_IO.Put (Ascii.HT);
				Ada.Text_IO.Put_Line (Page_Info_Maps.Key (I));
				Page_Info_Maps.Next (I);
			end loop;
		end;
	end if;
	-- build pages
	declare
		I : Page_Info_Maps.Cursor := Page_Info.First;
	begin
		while Page_Info_Maps.Has_Element (I) loop
			declare
				Path : constant String := Page_Info_Maps.Key (I);
				Source : constant String :=
					Ada.Directories.Compose (Name => Path, Extension => Pages.Source_Extension);
				Info : Pages.Page_Info_Access
					renames Page_Info.Constant_Reference (I);
			begin
				if Ada.Directories.Modification_Time (Source) > Info.Build_Time then
					if Verbose then
						Ada.Text_IO.Put (Path);
						if Info.Is_New then
							Ada.Text_IO.Put (": new file.");
						else
							Ada.Text_IO.Put (": modified.");
						end if;
						Ada.Text_IO.New_Line;
					end if;
					Pages.Page_Sets.Insert (Modified_Pages, Path);
					Pages.Page_Sets.Insert (Build_Pages, Path);
				elsif Force then
					if Verbose then
						Ada.Text_IO.Put (Path);
						Ada.Text_IO.Put (": forced.");
						Ada.Text_IO.New_Line;
					end if;
					Pages.Page_Sets.Insert (Build_Pages, Path);
				end if;
			end;
			Page_Info_Maps.Next (I);
		end loop;
	end;
	declare
		Handled : Boolean := True;
	begin
		while Handled loop
			Handled := False;
			declare
				I : Page_Info_Maps.Cursor := Page_Info.First;
			begin
				while Page_Info_Maps.Has_Element (I) loop
					declare
						Path : constant String := Page_Info_Maps.Key (I);
					begin
						if not Build_Pages.Contains (Path) then
							declare
								Page_Info_Item : Pages.Page_Info_Access
									renames Page_Info.Constant_Reference (I);
								Build_Time : constant Ada.Calendar.Time := Page_Info_Item.Build_Time;
								J : Pages.Page_Sets.Cursor := Page_Info_Item.Reference.First;
							begin
								Deps: while Pages.Page_Sets.Has_Element (J) loop
									declare
										Ref_Path : constant String :=
											Pages.Reference_Path (
												Pages.Compose (
													Pages.Containing_Directory (Path),
													Page_Info_Item.Reference.Constant_Reference (
														J)));
										Adding : Boolean := False;
									begin
										if Ref_Path (Ref_Path'Last) = '/' then
											null;
										elsif Ada.Directories.Exists (Ref_Path)
											and then not Ada.Directories.Exists (
												Ada.Directories.Compose (
													Name => Ref_Path,
													Extension => Pages.Source_Extension))
										then
											if Ada.Directories.Modification_Time (Ref_Path) > Build_Time then
												Adding := True;
											end if;
										elsif Build_Pages.Contains (Ref_Path) then
											Adding := True;
										else
											if not Page_Info.Contains (Ref_Path) then
												Page_Info_Maps.Insert (
													Page_Info,
													Ref_Path,
													new Pages.Page_Info'(Pages.Get_Page_Info (Ref_Path)));
											end if;
											if Page_Info.Constant_Reference (Ref_Path).Build_Time >
												Build_Time
											then
												Adding := True;
											end if;
										end if;
										if Adding then
											if Verbose then
												Ada.Text_IO.Put (Path);
												Ada.Text_IO.Put (": ");
												Ada.Text_IO.Put (Ref_Path);
												Ada.Text_IO.New_Line;
											end if;
											Pages.Page_Sets.Insert (Build_Pages, Path);
											Handled := True;
											exit Deps;
										end if;
									end;
									Pages.Page_Sets.Next (J);
								end loop Deps;
							end;
						end if;
					end;
					Page_Info_Maps.Next (I);
				end loop;
			end;
		end loop;
	end;
	if Verbose then
		Ada.Text_IO.Put_Line ("build pages:");
		declare
			I : Pages.Page_Sets.Cursor := Build_Pages.First;
		begin
			while Pages.Page_Sets.Has_Element (I) loop
				Ada.Text_IO.Put (Ascii.HT);
				Ada.Text_IO.Put_Line (Build_Pages.Constant_Reference (I));
				Pages.Page_Sets.Next (I);
			end loop;
		end;
	end if;
	-- parse and output
	declare
		Build_Time : constant Ada.Calendar.Time := Ada.Calendar.Clock;
		Parsed_Pages : Pages.Page_Sets.Set;
		procedure Process (Path : in String) is
		begin
			if not Parsed_Pages.Contains (Path) then
				declare
					Source_Path : constant String :=
						Ada.Directories.Compose (Name => Path, Extension => Pages.Source_Extension);
					Info : Pages.Page_Info_Access :=
						Page_Info.Constant_Reference (Path);
					File : Ada.Streams.Stream_IO.File_Type;
					Current_Page_Contents : Markup.Page_Access := new Markup.Page;
				begin
					Ada.Text_IO.Put ("parsing ");
					Ada.Text_IO.Put (Source_Path);
					Ada.Text_IO.Put ("...");
					Page_Maps.Insert (Page_Contents, Path, Current_Page_Contents);
					Info.Build_Time := Build_Time;
					Info.Reference := Pages.Page_Sets.Empty_Set;
					Ada.Streams.Stream_IO.Open (File, Ada.Streams.Stream_IO.In_File,
						Name => Source_Path);
					declare
						Size : constant Ada.Streams.Stream_IO.Count :=
							Ada.Streams.Stream_IO.Size (File);
						Source : String (1 .. Natural (Size));
						Errors : Markup.Error_Lists.List;
					begin
						String'Read (Ada.Streams.Stream_IO.Stream (File), Source);
						Ada.Streams.Stream_IO.Close (File);
						Parse (
							Source_Path,
							Source,
							Current_Page_Contents.all,
							Info.Reference'Access,
							Errors => Errors);
						Error := not Errors.Is_Empty;
						if not Error then
							Ada.Text_IO.Put_Line ("ok");
							Pages.Page_Sets.Insert (Parsed_Pages, Path);
							declare
								I : Pages.Page_Sets.Cursor := Info.Reference.First;
							begin
								while Pages.Page_Sets.Has_Element (I) loop
									declare
										Target_Path_From_Page : constant String :=
											Info.Reference.Constant_Reference (I);
										Target_Path : constant String :=
											Pages.Reference_Path (
												Pages.Compose (
													Pages.Containing_Directory (Source_Path),
													Target_Path_From_Page));
									begin
										if Verbose then
											Ada.Text_IO.Put (Ascii.HT);
											Ada.Text_IO.Put (Path);
											Ada.Text_IO.Put (": ");
											Ada.Text_IO.Put (Target_Path);
											Ada.Text_IO.New_Line;
										end if;
										if Target_Path (Target_Path'Last) = '/' then
											null;
										elsif Ada.Directories.Exists (
											Ada.Directories.Compose (
												Name => Target_Path,
												Extension => Pages.Source_Extension))
										then
											if not Page_Info.Contains (Target_Path) then
												Page_Info_Maps.Insert (
													Page_Info,
													Target_Path,
													new Pages.Page_Info'(Pages.Get_Page_Info (Target_Path)));
											end if;
											Process (Target_Path);
											exit when Error;
										elsif Ada.Directories.Exists (Target_Path) then
											null;
										else
											Ada.Text_IO.Put (Target_Path);
											Ada.Text_IO.Put (": not found!");
											Ada.Text_IO.New_Line;
											Error := True;
											exit;
										end if;
									end;
									Pages.Page_Sets.Next (I);
								end loop;
							end;
							if not Error and Build_Pages.Contains (Path) then
								declare
									Dest_Path : constant String := Format.Output_Path (Path);
									Dest_File : Ada.Streams.Stream_IO.File_Type;
									procedure Output_String (S : in String) is
									begin
										String'Write (Ada.Streams.Stream_IO.Stream (Dest_File), S);
									end Output_String;
									procedure Navigate (Process : not null access procedure (Path : in String)) is
										Containing_Directory : constant String := Pages.Containing_Directory (Path);
										function Parent_Directory return String is
										begin
											if Containing_Directory = "" or else Containing_Directory = "." then
												return "..";
											else
												return Pages.Containing_Directory (Containing_Directory);
											end if;
										end Parent_Directory;
										Parent_Path : constant String :=
											Ada.Directories.Compose (
												Containing_Directory => Parent_Directory,
												Name => Pages.Index);
										Parent_Source : constant String :=
											Ada.Directories.Compose (
												Name => Parent_Path,
												Extension => Pages.Source_Extension);
										Index_Path : constant String :=
											Ada.Directories.Compose (
												Containing_Directory => Containing_Directory,
												Name => Pages.Index);
										Index_Source : constant String :=
											Ada.Directories.Compose (
												Name => Index_Path,
												Extension => Pages.Source_Extension);
									begin
										if Ada.Directories.Exists (Parent_Source) then
											Process (Parent_Path);
										end if;
										if Path /= Index_Path and then Ada.Directories.Exists (Index_Source) then
											Process (Index_Path);
										end if;
									end Navigate;
									procedure Query (
										Path : in String;
										Process : not null access procedure (Page : in Markup.Page)) is
									begin
										Process (
											Page_Contents.Constant_Reference (Path).all);
									end Query;
								begin
									Ada.Text_IO.Put ("generating ");
									Ada.Text_IO.Put (Dest_Path);
									Ada.Text_IO.Put ("...");
									Ada.Streams.Stream_IO.Create (Dest_File, Name => Dest_Path);
									Output.Output (
										Format.all,
										Dest_Path,
										Current_Page_Contents.all,
										Navigate => Navigate'Access,
										Query => Query'Access,
										Output => Output_String'Access);
									Ada.Streams.Stream_IO.Close (Dest_File);
									Pages.Save_Page_Info (Path, Info.all, Source, Modified_Pages.Contains (Path));
									Ada.Text_IO.Put_Line ("ok");
								end;
							end if;
						else
							Ada.Text_IO.Put_Line ("error!");
							declare
								I : Markup.Error_Lists.Cursor := Errors.First;
							begin
								while Markup.Error_Lists.Has_Element (I) loop
									declare
										Item : Markup.Error_Item
											renames Errors.Constant_Reference (I);
									begin
										Ada.Text_IO.Put (Ada.Strings.Unbounded.To_String (Item.Position.File_Name));
										Ada.Text_IO.Put (':');
										Ada.Text_IO.Put (To_String (Item.Position.Line));
										Ada.Text_IO.Put (':');
										Ada.Text_IO.Put (To_String (Item.Position.Column));
										Ada.Text_IO.Put (": ");
										Ada.Text_IO.Put (Ada.Strings.Unbounded.To_String (Item.Message));
										Ada.Text_IO.New_Line;
									end;
									Markup.Error_Lists.Next (I);
								end loop;
							end;
						end if;
					end;
				end;
			end if;
		end Process;
		I : Pages.Page_Sets.Cursor := Build_Pages.First;
	begin
		while Pages.Page_Sets.Has_Element (I) loop
			Process (Build_Pages.Constant_Reference (I));
			exit when Error;
			Pages.Page_Sets.Next (I);
		end loop;
	end;
	-- free memories for page info
	declare
		I : Page_Info_Maps.Cursor := Page_Info.First;
	begin
		while Page_Info_Maps.Has_Element (I) loop
			Pages.Free (Page_Info.Reference (I));
			Page_Info_Maps.Next (I);
		end loop;
	end;
	-- free memories for page contents
	declare
		I : Page_Maps.Cursor := Page_Contents.First;
	begin
		while Page_Maps.Has_Element (I) loop
			Markup.Free (Page_Contents.Reference (I));
			Page_Maps.Next (I);
		end loop;
	end;
end Scream.Build;
