with Ada.Containers.Indefinite_Ordered_Maps;
with Ada.Characters.Handling;
with Ada.Unchecked_Deallocation;
package body Scream.Markup is
	use type Ada.Tags.Tag;
	
	type Tag_With_Params is record
		Tag : Ada.Tags.Tag;
		Params : aliased Params_Type;
	end record;
	
	package Tag_Maps is new Ada.Containers.Indefinite_Ordered_Maps (
		String,
		Tag_With_Params);
	
	Parsers : Tag_Maps.Map;
	
	function Parser (Key : String) return Mark'Class is
		TwP : Tag_With_Params := Parsers.Constant_Reference (Key);
	begin
		return Create (TwP.Tag, TwP.Params'Access);
	end Parser;
	
	procedure Raw_Free is
		new Ada.Unchecked_Deallocation (Element'Class, Element_Access);
	procedure Raw_Free is new Ada.Unchecked_Deallocation (Page, Page_Access);
	
	function Slice (Page : Markup.Page; R : Range_Type) return String is
	begin
		return Ada.Strings.Unbounded.Slice (Page.Text, R.First, R.Last);
	end Slice;
	
	procedure Free (Page : in out Page_Access) is
		I : Element_Access;
	begin
		I := Page.All_Elements;
		while I /= null loop
			declare
				Next : Element_Access := I.Next_In_All_Elements;
			begin
				Raw_Free (I);
				I := Next;
			end;
		end loop;
		Raw_Free (Page);
	end Free;
	
	function Is_Empty (Item : Element) return Boolean is
		Found : exception;
		procedure Process (Item : in Element'Class) is
		begin
			raise Found;
		end Process;
	begin
		if Item.Contents.First <= Item.Contents.Last then
			return False;
		else
			begin
				Iterate (Item, Process'Access);
			exception
				when Found =>
					return False;
			end;
			return True;
		end if;
	end Is_Empty;
	
	procedure Iterate (
		Contents : Range_Type;
		Items : Inline_Access;
		Process_Text : not null access procedure (Item : in Range_Type);
		Process_Inline : not null access procedure (Item : in Inline'Class))
	is
		Index : Positive := Contents.First;
		I : Inline_Access := Items;
	begin
		while I /= null loop
			if Index < I.Contents.First then
				Process_Text ((First => Index, Last => I.Contents.First - 1));
			end if;
			Process_Inline (I.all);
			Index := I.Contents.Last + 1;
			I := I.Next;
		end loop;
		if Index <= Contents.Last then
			Process_Text ((First => Index, Last => Contents.Last));
		end if;
	end Iterate;
	
	procedure Iterate (
		Items : in Inline_Access;
		Process : not null access procedure (Item : in Element'Class))
	is
		I : Inline_Access := Items;
	begin
		while I /= null loop
			Process (I.all);
			I := I.Next;
		end loop;
	end Iterate;
	
	procedure Iterate (
		Items : in Block_Access;
		Process : not null access procedure (Item : in Block'Class))
	is
		I : Block_Access := Items;
	begin
		while I /= null loop
			Process (I.all);
			I := I.Next;
		end loop;
	end Iterate;
	
	procedure Iterate (
		Items : in Block_Access;
		Process : not null access procedure (Item : in Element'Class))
	is
		I : Block_Access := Items;
	begin
		while I /= null loop
			Process (I.all);
			I := I.Next;
		end loop;
	end Iterate;
	
	procedure Next (S : in String; P : in out Source_Position) is
	begin
		if S (P.Index) = Ascii.LF then
			P.Index := P.Index + 1;
			P.Line := P.Line + 1;
			P.Column := 1;
		elsif S (P.Index) = Ascii.CR then
			P.Index := P.Index + 1;
			P.Line := P.Line + 1;
			P.Column := 1;
			if S (P.Index) = Ascii.LF then
				P.Index := P.Index + 1;
			end if;
		else
			P.Index := P.Index + 1;
			P.Column := P.Column + 1;
		end if;
	end Next;
	
	procedure Skip_Spaces (S : in String; P : in out Source_Position) is
	begin
		while P.Index <= S'Last
			and then (S (P.Index) = ' ' or else S (P.Index) = Ascii.HT)
		loop
			P.Index := P.Index + 1;
			P.Column := P.Column + 1;
		end loop;
	end Skip_Spaces;
	
	procedure Skip_Single_Line_Feed (S : in String; P : in out Source_Position) is
	begin
		if P.Index <= S'Last
			and then (S (P.Index) = Ascii.LF or else S (P.Index) = Ascii.CR)
		then
			Next (S, P);
		end if;
	end Skip_Single_Line_Feed;
	
	procedure Take_Word (
		S : in String;
		Last : out Natural;
		P : in out Source_Position) is
	begin
		Last := P.Index - 1;
		while P.Index <= S'Last
			and then (S (P.Index) in 'A' .. 'Z' or else S (P.Index) in 'a' .. 'z')
		loop
			P.Index := P.Index + 1;
			P.Column := P.Column + 1;
			Last := Last + 1;
		end loop;
	end Take_Word;
	
	procedure Take_Line (
		S : in String;
		Last : out Natural;
		P : in out Source_Position) is
	begin
		Last := P.Index - 1;
		while P.Index <= S'Last
			and then S (P.Index) /= Ascii.LF
			and then S (P.Index) /= Ascii.CR
		loop
			P.Index := P.Index + 1;
			P.Column := P.Column + 1;
			Last := Last + 1;
		end loop;
		if P.Index <= S'Last and then S (P.Index) = Ascii.LF then
			P.Index := P.Index + 1;
			P.Line := P.Line + 1;
			P.Column := 1;
		elsif P.Index <= S'Last and then S (P.Index) = Ascii.CR then
			P.Index := P.Index + 1;
			P.Line := P.Line + 1;
			P.Column := 1;
			if P.Index <= S'Last and then S (P.Index) = Ascii.LF then
				P.Index := P.Index + 1;
			end if;
		end if;
	end Take_Line;
	
	procedure Take_Same_Symbols (
		S : in String;
		Symbol : in Character;
		Count : out Natural;
		P : in out Source_Position) is
	begin
		Count := 0;
		while P.Index < S'Last
			and then S (P.Index) = Symbol
		loop
			P.Index := P.Index + 1;
			P.Column := P.Column + 1;
			Count := Count + 1;
		end loop;
	end Take_Same_Symbols;
	
	procedure Error (
		Parsing : not null access Markup.Parsing;
		Position: in Source_Position;
		Message : in String) is
	begin
		Error_Lists.Append (
			Parsing.Errors,
			(Position, Ada.Strings.Unbounded.To_Unbounded_String (Message)));
	end Error;
	
	procedure Append (
		Parsing : not null access Markup.Parsing;
		Previous : in out Inline_Access_Access;
		Item : not null access Inline) is
	begin
		Item.Next_In_All_Elements := Parsing.All_Elements.all;
		Parsing.All_Elements.all := Element_Access (Item);
		Previous.all := Inline_Access (Item);
		Previous := Item.Next'Access;
	end Append;
	
	procedure Append (
		Parsing : not null access Markup.Parsing;
		Previous : in out Block_Access_Access;
		Item : not null access Block) is
	begin
		Item.Next_In_All_Elements := Parsing.All_Elements.all;
		Parsing.All_Elements.all := Element_Access (Item);
		Previous.all := Block_Access (Item);
		Previous := Item.Next'Access;
	end Append;
	
	procedure Append (
		Parsing : not null access Markup.Parsing;
		R : out Range_Type;
		S : in String) is
	begin
		R.First := Ada.Strings.Unbounded.Length (Parsing.Text.all) + 1;
		Ada.Strings.Unbounded.Append (Parsing.Text.all, S);
		R.Last := Ada.Strings.Unbounded.Length (Parsing.Text.all);
	end Append;
	
	procedure Parse_Inline (
		Parsing : not null access Markup.Parsing;
		Previous : in Inline_Access_Access;
		Source : String;
		Position : in out Source_Position;
		Contents : out Range_Type;
		Stopper : not null access function (Source : String) return Boolean)
	is
		Sub_Contents : Range_Type;
		Variable_Previous : Inline_Access_Access := Previous;
	begin
		Append (Parsing, Contents, "");
		while not Stopper (Source (Position.Index .. Source'Last)) loop
			if Source (Position.Index) = '[' then
				declare
					Paren : String (1 .. 16);
					Paren_Length : Natural := 0;
					function Is_Closed_Paren (S : String) return Boolean is
						Index : Positive := S'First;
					begin
						if Is_End_Of_Line (S) then
							return True;
						end if;
						for I in reverse 1 .. Paren_Length loop
							if Index > Source'Last or else Source (Index) /= Paren (I) then
								return False;
							end if;
							Index := Index + 1;
						end loop;
						return True;
					end Is_Closed_Paren;
					Mark_First : Source_Position;
					Mark_Last : Natural;
					Argument_First : Source_Position;
					Argument_Last : Natural;
					Has_Argument : Boolean;
				begin
					while Position.Index <= Source'Last loop
						case Source (Position.Index) is
							when '[' =>
								Paren_Length := Paren_Length + 1;
								Paren (Paren_Length) := ']';
								Next (Source, Position);
							when '(' =>
								Paren_Length := Paren_Length + 1;
								Paren (Paren_Length) := ')';
								Next (Source, Position);
							when '{' =>
								Paren_Length := Paren_Length + 1;
								Paren (Paren_Length) := '}';
								Next (Source, Position);
							when '<' =>
								Paren_Length := Paren_Length + 1;
								Paren (Paren_Length) := '>';
								Next (Source, Position);
							when others =>
								exit;
						end case;
					end loop;
					Skip_Spaces (Source, Position);
					if Position.Index <= Source'Last and then Source (Position.Index) = '.' then
						Mark_First := Position;
						Mark_Last := Position.Index;
						Has_Argument := True;
					elsif Position.Index <= Source'Last and then Source (Position.Index) = '!' then
						Mark_First := Position;
						Mark_Last := Position.Index;
						Next (Source, Position);
						Skip_Spaces (Source, Position);
						Has_Argument := True;
					else
						Mark_First := Position;
						Take_Word (Source, Mark_Last, Position);
						Skip_Spaces (Source, Position);
						Has_Argument :=
							Position.Index <= Source'Last and then Source (Position.Index) = ':';
						if Has_Argument then
							Next (Source, Position);
							Skip_Spaces (Source, Position);
						end if;
					end if;
					if Has_Argument then
						Argument_First := Position;
						Argument_Last := Position.Index - 1;
						while not Is_Closed_Paren (Source (Position.Index .. Source'Last))
							and then Source (Position.Index) /= ';'
						loop
							Argument_Last := Position.Index;
							Next (Source, Position);
						end loop;
					else
						Argument_First := Position;
						Argument_Last := Position.Index - 1;
					end if;
					if Position.Index <= Source'Last and then Source (Position.Index) = ';' then
						Next (Source, Position);
						Skip_Spaces (Source, Position);
					end if;
					declare
						Mark_Name : constant String :=
							Ada.Characters.Handling.To_Upper (Source (Mark_First.Index .. Mark_Last));
					begin
						if Parsers.Contains (Mark_Name) then
							declare
								The_Markup : constant Mark'Class := Parser (Mark_Name);
							begin
								Parse_Inline (
									The_Markup,
									Source (Argument_First.Index .. Argument_Last),
									Argument_First,
									Parsing,
									Variable_Previous,
									Source (Position.Index .. Source'Last),
									Position,
									Is_Closed_Paren'Access);
							end;
						else
							Error (Parsing, Mark_First, Mark_Name & " is undefined function.");
							while not Is_Closed_Paren (Source (Position.Index .. Source'Last)) loop
								Next (Source, Position);
							end loop;
						end if;
					end;
					if Is_End_Of_Line (Source (Position.Index .. Source'Last)) then
						Error (Parsing, Position, "parenthesis was mismatched.");
					else
						for I in 1 .. Paren_Length loop
							Next (Source, Position);
						end loop;
					end if;
				end;
			elsif Source (Position.Index) = ']' then
				Error (Parsing, Position, "parenthesis was mismatched.");
				Next (Source, Position);
			elsif Source (Position.Index) = '\' then
				Next (Source, Position);
				if Position.Index <= Source'Last then
					if Source (Position.Index) = Ascii.LF
						or else Source (Position.Index) = Ascii.CR
					then
						Next (Source, Position);
					else
						Append (Parsing, Sub_Contents, Source (Position.Index .. Position.Index));
						Contents.Last := Sub_Contents.Last;
						Next (Source, Position);
					end if;
				end if;
			else
				Append (Parsing, Sub_Contents, Source (Position.Index .. Position.Index));
				Contents.Last := Sub_Contents.Last;
				Next (Source, Position);
			end if;
		end loop;
	end Parse_Inline;
	
	procedure Parse_Block (
		Parsing : not null access Markup.Parsing;
		Previous : in Block_Access_Access;
		Source : String;
		Position : in out Source_Position;
		Contents : out Range_Type;
		Stopper : not null access function (Source : String) return Boolean)
	is
		Level : Natural;
		function Is_Closed_By_Level (S : String) return Boolean is
			Count : Natural := 0;
			Index : Positive := S'First;
			Result : Boolean;
		begin
			if S'Length = 0 then
				Result := True;
			else
				while Index <= S'Last and then S (Index) = '-' loop
					Count := Count + 1;
					Index := Index + 1;
				end loop;
				Result := Count = Level
					and then (
						Index > S'Last
						or else S (Index) = Ascii.LF
						or else S (Index) = Ascii.CR);
			end if;
			return Result;
		end Is_Closed_By_Level;
		Variable_Previous : Block_Access_Access := Previous;
		Last : Natural;
		Next_First : Source_Position;
		Mark_First : Source_Position;
		Mark_Last : Natural;
		Argument_First : Source_Position;
		Argument_Last : Natural;
	begin
		Contents.First := Ada.Strings.Unbounded.Length (Parsing.Text.all) + 1;
		while not Stopper (Source (Position.Index .. Source'Last)) loop
			Next_First := Position;
			Take_Line (Source, Last, Next_First);
			if Last < Position.Index then
				Position := Next_First;
			elsif Source (Position.Index) = '=' then
				Take_Same_Symbols (Source, '=', Level, Position);
				Skip_Spaces (Source, Position);
				Mark_First := Position;
				Take_Word (Source, Mark_Last, Position);
				Skip_Spaces (Source, Position);
				if Position.Index <= Source'Last and then Source (Position.Index) = ':' then
					Next (Source, Position);
					Skip_Spaces (Source, Position);
					Argument_First := Position;
					Argument_Last := Position.Index - 1;
					while Position.Index <= Last and then Source (Position.Index) /= ';' loop
						Argument_Last := Position.Index;
						Next (Source, Position);
					end loop;
				else
					Argument_First := Position;
					Argument_Last := Position.Index - 1;
				end if;
				if Position.Index <= Source'Last and then Source (Position.Index) = ';' then
					Next (Source, Position);
					Skip_Spaces (Source, Position);
				end if;
				declare
					Mark_Name : constant String :=
						Ada.Characters.Handling.To_Upper (Source (Mark_First.Index .. Mark_Last));
				begin
					if Parsers.Contains (Mark_Name) then
						declare
							Base : constant Positive := Position.Index;
							function Is_After (S : String) return Boolean is
							begin
								return S'First > S'Last or else S'First > Base;
							end Is_After;
							The_Markup : constant Mark'Class := Parser (Mark_Name);
						begin
							Parse_Block (
								The_Markup,
								Source (Argument_First.Index .. Argument_Last),
								Argument_First,
								Parsing,
								Variable_Previous,
								Source (Position.Index .. Source'Last),
								Position,
								Is_After'Access);
						end;
					else
						Error (Parsing, Mark_First, Mark_Name & " is undefined function.");
					end if;
				end;
			elsif Position.Index + 1 <= Source'Last
				and then Source (Position.Index) = '-'
				and then Source (Position.Index + 1) = '-'
			then
				Take_Same_Symbols (Source, '-', Level, Position);
				Skip_Spaces (Source, Position);
				Mark_First := Position;
				if Source (Position.Index) = '!' then
					Mark_Last := Position.Index;
					Next (Source, Position);
					Skip_Spaces (Source, Position);
					Argument_First := Position;
					Argument_Last := Last;
				else
					Take_Word (Source, Mark_Last, Position);
					Skip_Spaces (Source, Position);
					if Position.Index <= Source'Last and then Source (Position.Index) = ':' then
						Next (Source, Position);
						Skip_Spaces (Source, Position);
						Argument_First := Position;
						Argument_Last := Last;
					else
						Argument_First := Position;
						Argument_Last := Position.Index - 1;
					end if;
				end if;
				Position := Next_First;
				declare
					Mark_Name : constant String :=
						Ada.Characters.Handling.To_Upper (Source (Mark_First.Index .. Mark_Last));
				begin
					if Parsers.Contains (Mark_Name) then
						declare
							The_Markup : constant Mark'Class := Parser (Mark_Name);
						begin
							Parse_Block (
								The_Markup,
								Source (Argument_First.Index .. Argument_Last),
								Argument_First,
								Parsing,
								Variable_Previous,
								Source (Position.Index .. Source'Last),
								Position,
								Is_Closed_By_Level'Access);
						end;
					else
						Error (Parsing, Mark_First, Mark_Name & " is undefined function.");
						while not Is_Closed_By_Level (Source (Position.Index .. Source'Last)) loop
							Take_Line (Source, Last, Position);
						end loop;
					end if;
				end;
				if Position.Index > Source'Last then
					Error (Parsing, Position, "block was mismatched.");
				else
					Skip_Single_Line_Feed (Source, Position);
					Take_Line (Source, Last, Position);
				end if;
			elsif Source (Position.Index) = '-' then
				Argument_First := Position;
				declare
					function Is_Closed_By_Not_Item (S : String) return Boolean is
					begin
						return S'First > S'Last
							or else (
								S (S'First) /= '-'
								and then S (S'First) /= ' '
								and then S (S'First) /= Ascii.HT)
							or else Stopper (S);
					end Is_Closed_By_Not_Item;
					List_Mark : constant Mark'Class := Parser ("LIST");
				begin
					Parse_Block (
						List_Mark,
						"",
						Argument_First,
						Parsing,
						Variable_Previous,
						Source (Position.Index .. Source'Last),
						Position,
						Is_Closed_By_Not_Item'Access);
				end;
			elsif Source (Position.Index) = '|' then
				Argument_First := Position;
				declare
					function Is_Closed_By_Not_Item (S : String) return Boolean is
					begin
						return S'First > S'Last or else S (S'First) /= '|' or else Stopper (S);
					end Is_Closed_By_Not_Item;
					Table : constant Mark'Class := Parser ("TABLE");
				begin
					Parse_Block (
						Table,
						"",
						Argument_First,
						Parsing,
						Variable_Previous,
						Source (Position.Index .. Source'Last),
						Position,
						Is_Closed_By_Not_Item'Access);
				end;
			elsif Source (Position.Index) = '*' then
				declare
					function Is_Closed_By_Next_Section (S : String) return Boolean is
					begin
						return S'First > S'Last or else S (S'First) = '*' or else Stopper (S);
					end Is_Closed_By_Next_Section;
					Section : constant Mark'Class := Parser ("SECTION");
				begin
					Next (Source, Position);
					Skip_Spaces (Source, Position);
					Argument_First := Position;
					Argument_Last := Last;
					Position := Next_First;
					if Position.Index + 1 <= Source'Last
						and then Source (Position.Index) = '-'
						and then Source (Position.Index + 1) = '-'
					then
						Take_Same_Symbols (Source, '-', Level, Position);
						Take_Line (Source, Last, Position);
						Parse_Block (
							Section,
							Source (Argument_First.Index .. Argument_Last),
							Argument_First,
							Parsing,
							Variable_Previous,
							Source (Position.Index .. Source'Last),
							Position,
							Is_Closed_By_Level'Access);
						Skip_Single_Line_Feed (Source, Position);
						Take_Line (Source, Last, Position);
					else
						Parse_Block (
							Section,
							Source (Argument_First.Index .. Argument_Last),
							Argument_First,
							Parsing,
							Variable_Previous,
							Source (Position.Index .. Source'Last),
							Position,
							Is_Closed_By_Next_Section'Access);
					end if;
				end;
			else
				Argument_First := Position;
				declare
					function Is_Closed_By_Blank_Line (S : String) return Boolean is
					begin
						return Is_End_Of_Line (S)
							or else S (S'First) = '='
							or else S (S'First) = '-'
							or else S (S'First) = '|'
							or else S (S'First) = '*'
							or else Stopper (S);
					end Is_Closed_By_Blank_Line;
					P : constant Mark'Class := Parser ("P");
				begin
					Parse_Block (
						P,
						"",
						Argument_First,
						Parsing,
						Variable_Previous,
						Source (Position.Index .. Source'Last),
						Position,
						Is_Closed_By_Blank_Line'Access);
				end;
			end if;
		end loop;
		Contents.Last := Ada.Strings.Unbounded.Length (Parsing.Text.all);
	end Parse_Block;
	
	procedure Parse_Inline (
		The_Mark : in Mark;
		Argument : in String;
		Argument_Position : in Source_Position;
		Parsing : not null access Markup.Parsing;
		Previous : in out Inline_Access_Access;
		Source : in String;
		Position : in out Source_Position;
		Stopper : not null access function (Source : String) return Boolean) is
	begin
		Error (Parsing, Position, "this function can not be used as inline.");
		while not Stopper (Source (Position.Index .. Source'Last)) loop
			Next (Source, Position);
		end loop;
	end Parse_Inline;
	
	procedure Parse_Block (
		The_Mark : in Mark;
		Argument : in String;
		Argument_Position : in Source_Position;
		Parsing : not null access Markup.Parsing;
		Previous : in out Block_Access_Access;
		Source : in String;
		Position : in out Source_Position;
		Stopper : not null access function (Source : String) return Boolean)
	is
		Last : Natural;
	begin
		Error (Parsing, Position, "this function can not be used as block.");
		while not Stopper (Source (Position.Index .. Source'Last)) loop
			Take_Line (Source, Last, Position);
		end loop;
	end Parse_Block;
	
	procedure Register (
		Name : in String;
		Tag : in Ada.Tags.Tag;
		Params : in Params_Type := (Data => 0)) is
	begin
		Tag_Maps.Insert (Parsers, Name, (Tag, Params));
	end Register;
	
	function Is_Empty (S : String) return Boolean is
	begin
		return S'Length = 0;
	end Is_Empty;
	
	function Is_End_Of_Line (S : String) return Boolean is
	begin
		return S'Length = 0
			or else S (S'First) = Ascii.LF
			or else S (S'First) = Ascii.CR;
	end Is_End_Of_Line;
	
	function Generic_Inline_Stopper (S : String) return Boolean is
		Index : Positive := S'First;
	begin
		if Index > S'Last then
			return True;
		elsif S (Index) = Ascii.LF then
			return Block_Stopper (S (Index + 1 .. S'Last));
		elsif S (Index) = Ascii.CR then
			Index := Index + 1;
			if Index <= S'Last and then S (Index) = Ascii.LF then
				Index := Index + 1;
			end if;
			return Block_Stopper (S (Index .. S'Last));
		else
			return False;
		end if;
	end Generic_Inline_Stopper;
	
end Scream.Markup;
