with Ada.Containers.Doubly_Linked_Lists;
with Ada.Tags.Generic_Dispatching_Constructor;
with Ada.Strings.Unbounded;
with Scream.Pages;
package Scream.Markup is
	
	-- Document
	
	type Range_Type is record
		First : Positive;
		Last : Natural;
	end record;
	
	type Element;
	type Element_Access is access all Element'Class;
	type Element is abstract tagged limited record
		Next_In_All_Elements : Element_Access;
		Contents : Range_Type;
	end record;
	
	function Is_Empty (Item : Element) return Boolean;
	
	procedure Iterate (
		Item : in Element;
		Process : not null access procedure (Item : in Element'Class)) is null;
	
	type Inline is tagged;
	type Inline_Access is access constant Inline'Class;
	type Inline is abstract new Element with record
		Next : aliased Inline_Access;
	end record;
	
	type Block is tagged;
	type Block_Access is access constant Block'Class;
	type Block is abstract new Element with record
		Next : aliased Block_Access;
	end record;
	
	type Page is tagged limited record
		Text : aliased Ada.Strings.Unbounded.Unbounded_String;
		Title : Range_Type;
		Items : aliased Block_Access;
		All_Elements : aliased Element_Access;
	end record;
	
	function Slice (Page : Markup.Page; R : Range_Type) return String;
	
	type Page_Access is access Page;
	
	procedure Free (Page : in out Page_Access);
	
	procedure Iterate (
		Contents : in Range_Type;
		Items : in Inline_Access;
		Process_Text : not null access procedure (Item : in Range_Type);
		Process_Inline : not null access procedure (Item : in Inline'Class));
	procedure Iterate (
		Items : in Inline_Access;
		Process : not null access procedure (Item : in Element'Class));
	procedure Iterate (
		Items : in Block_Access;
		Process : not null access procedure (Item : in Block'Class));
	procedure Iterate (
		Items : in Block_Access;
		Process : not null access procedure (Item : in Element'Class));
	
	-- Parsing
	
	type Source_Position is record
		File_Name : Ada.Strings.Unbounded.Unbounded_String;
		Index : Natural;
		Line : Positive;
		Column : Positive;
	end record;
	
	procedure Next (S : in String; P : in out Source_Position);
	procedure Skip_Spaces (S : in String; P : in out Source_Position);
	procedure Skip_Single_Line_Feed (S : in String; P : in out Source_Position);
	procedure Take_Word (
		S : in String;
		Last : out Natural;
		P : in out Source_Position);
	procedure Take_Line (
		S : in String;
		Last : out Natural;
		P : in out Source_Position);
	procedure Take_Same_Symbols (
		S : in String;
		Symbol : in Character;
		Count : out Natural;
		P : in out Source_Position);
	
	type Inline_Access_Access is not null access all Inline_Access;
	type Block_Access_Access is not null access all Block_Access;
	
	type Error_Item is record
		Position : Source_Position;
		Message : Ada.Strings.Unbounded.Unbounded_String;
	end record;
	
	package Error_Lists is new Ada.Containers.Doubly_Linked_Lists (Error_Item);
	
	type Parsing is limited record
		Text : not null access Ada.Strings.Unbounded.Unbounded_String;
		All_Elements : not null access Element_Access;
		Reference : not null access Pages.Page_Sets.Set;
		Errors : Error_Lists.List;
	end record;
	
	procedure Error (
		Parsing : not null access Markup.Parsing;
		Position: in Source_Position;
		Message : in String);
	
	procedure Append (
		Parsing : not null access Markup.Parsing;
		Previous : in out Inline_Access_Access;
		Item : not null access Inline);
	
	procedure Append (
		Parsing : not null access Markup.Parsing;
		Previous : in out Block_Access_Access;
		Item : not null access Block);
	
	procedure Append (
		Parsing : not null access Markup.Parsing;
		R : out Range_Type;
		S : in String);
	
	procedure Parse_Inline (
		Parsing : not null access Markup.Parsing;
		Previous : in Inline_Access_Access;
		Source : in String;
		Position : in out Source_Position;
		Contents : out Range_Type;
		Stopper : not null access function (Source : String) return Boolean);
	
	procedure Parse_Block (
		Parsing : not null access Markup.Parsing;
		Previous : in Block_Access_Access;
		Source : in String;
		Position : in out Source_Position;
		Contents : out Range_Type;
		Stopper : not null access function (Source : String) return Boolean);
	
	-- mark
	
	type Params_Type is record
		Data : Integer;
	end record;
	
	type Mark is abstract tagged limited null record;
	
	procedure Parse_Inline (
		The_Mark : in Mark;
		Argument : in String;
		Argument_Position : in Source_Position;
		Parsing : not null access Markup.Parsing;
		Previous : in out Inline_Access_Access;
		Source : in String;
		Position : in out Source_Position;
		Stopper : not null access function (Source : String) return Boolean);
	
	procedure Parse_Block (
		The_Mark : in Mark;
		Argument : in String;
		Argument_Position : in Source_Position;
		Parsing : not null access Markup.Parsing;
		Previous : in out Block_Access_Access;
		Source : in String;
		Position : in out Source_Position;
		Stopper : not null access function (Source : String) return Boolean);
	
	function Create (Params : not null access Params_Type)
		return Mark is abstract;
	
	function Create is new Ada.Tags.Generic_Dispatching_Constructor (
		Mark, Params_Type, Create);
	
	procedure Register (
		Name : in String;
		Tag : in Ada.Tags.Tag;
		Params : in Params_Type := (Data => 0));
	
	-- stoppers
	
	function Is_Empty (S : String) return Boolean;
	function Is_End_Of_Line (S : String) return Boolean;
	
	generic
		with function Block_Stopper (S : String) return Boolean;
	function Generic_Inline_Stopper (S : String) return Boolean;
	
end Scream.Markup;
