with Ada.Calendar.Formatting;
with Ada.Calendar.Time_Zones;
with Ada.Directories;
with Ada.Strings.Unbounded;
with Web;
with Scream.Markup.Abbr;
with Scream.Markup.Anchor;
with Scream.Markup.Author;
with Scream.Markup.Code;
with Scream.Markup.Date;
with Scream.Markup.Delete;
with Scream.Markup.Emphasis;
with Scream.Markup.History;
with Scream.Markup.Http;
with Scream.Markup.Include;
with Scream.Markup.Input;
with Scream.Markup.List;
with Scream.Markup.Mailto;
with Scream.Markup.Output;
with Scream.Markup.Paragraph;
with Scream.Markup.Plain;
with Scream.Markup.Quote;
with Scream.Markup.Remark;
with Scream.Markup.Section;
with Scream.Markup.Subject;
with Scream.Markup.Table;
with Scream.Pages;
package body Scream.Output.HTML is
	use type Ada.Calendar.Time_Zones.Time_Offset;
	use type Ada.Strings.Unbounded.Unbounded_String;
	
	function To_String (X : Integer) return String is
		Image : constant String := Integer'Image (X);
	begin
		if Image (Image'First) = ' ' then
			return Image (Image'First + 1 .. Image'Last);
		else
			return Image;
		end if;
	end To_String;
	
	function Emphasis_Tag (Style : Markup.Emphasis.Style) return String is
	begin
		case Style is
			when Markup.Emphasis.Big =>
				return "big";
			when Markup.Emphasis.Bold =>
				return "b";
			when Markup.Emphasis.Cite =>
				return "cite";
			when Markup.Emphasis.Definition =>
				return "dfn";
			when Markup.Emphasis.Emphasis =>
				return "em";
			when Markup.Emphasis.Italic =>
				return "i";
			when Markup.Emphasis.Mark =>
				return "mark";
			when Markup.Emphasis.Small =>
				return "small";
			when Markup.Emphasis.Strong =>
				return "strong";
			when Markup.Emphasis.Sub =>
				return "sub";
			when Markup.Emphasis.Sup =>
				return "sup";
			when Markup.Emphasis.Variable =>
				return "var";
		end case;
	end Emphasis_Tag;
	
	overriding function Output_Path (Object : HTML; Path : String)
		return String is
	begin
		return Path & ".html";
	end Output_Path;
	
	overriding procedure Output (
		Object : in HTML;
		Path : in String;
		Page : in Markup.Page;
		Navigate : not null access procedure (
			Process : not null access procedure (Path : in String));
		Query : not null access procedure (
			Path : in String;
			Process : not null access procedure (Page : in Markup.Page));
		Output : not null access procedure (Item : in String))
	is
		Parent_Path : constant String := Pages.Containing_Directory (Path);
		function Get_Style_Sheet (The_Path : String) return String is
			Style_Sheet : constant String :=
				Ada.Directories.Compose (The_Path, "style.css");
		begin
			if Ada.Directories.Exists (Style_Sheet) then
				return Pages.Relative_Name (Parent_Path, Style_Sheet);
			elsif The_Path = ""
				or else (
					The_Path'First + 2 <= The_Path'Last
					and then The_Path (The_Path'First) = '.'
					and then The_Path (The_Path'First + 1) = '.'
					and then The_Path (The_Path'First + 2) = '/')
			then
				declare
					The_Parent : constant String := "../" & The_Path;
				begin
					if not Ada.Directories.Exists (
						Ada.Directories.Compose (The_Parent, Pages.Index, Pages.Source_Extension))
					then
						return "";
					else
						return Get_Style_Sheet (The_Parent);
					end if;
				end;
			else
				declare
					The_Parent : constant String := Pages.Containing_Directory (The_Path);
				begin
					if not Ada.Directories.Exists (
						Ada.Directories.Compose (The_Parent, Pages.Index, Pages.Source_Extension))
					then
						return "";
					else
						return Get_Style_Sheet (The_Parent);
					end if;
				end;
			end if;
		end Get_Style_Sheet;
		Style_Sheet : constant String := Get_Style_Sheet (Parent_Path);
		procedure Output_In_HTML is
			new Web.Generic_Write_In_HTML (Output.all, Web.XHTML);
		procedure Output_In_Attribute is
			new Web.Generic_Write_In_Attribute (Output.all, Web.XHTML);
		Mailto_Index : Natural;
		procedure Process_Mailto (Path : in String; Page : in Markup.Page) is
			Item_Exists : Boolean := False;
			procedure Process_Element (Item : in Markup.Element'Class) is
			begin
				if Item in Markup.Mailto.Inline then
					declare
						Mailto : Markup.Mailto.Inline renames Markup.Mailto.Inline (Item);
						Link : constant String := Ada.Strings.Unbounded.To_String (Mailto.Link);
					begin
						if not Item_Exists then
							Output ("<script>" & Ascii.LF);
							Output ("function set_m(){" & Ascii.LF);
							Output (Ascii.HT & "var d, s;" & Ascii.LF);
						end if;
						Mailto_Index := Mailto_Index + 1;
						Output (Ascii.HT & "d = [");
						for I in Link'Range loop
							if I > Link'First then
								Output (", ");
							end if;
							Output (To_String (Character'Pos (Link (I))));
						end loop;
						Output ("];" & Ascii.LF);
						Output (Ascii.HT & "s = """";" & Ascii.LF);
						Output (
							Ascii.HT
								& "for(var i = 0; i < d.length; ++i) s += String.fromCharCode(d[i]);"
								& Ascii.LF);
						Output (Ascii.HT & "document.getElementById(""m");
						Output (To_String (Mailto_Index));
						Output (""").href = s;" & Ascii.LF);
						Item_Exists := True;
					end;
				elsif Item in Markup.Include.Block then
					declare
						Include : Markup.Include.Block renames Markup.Include.Block (Item);
						Target_Path : constant String :=
							Pages.Compose (
								Pages.Containing_Directory (Path),
								Ada.Strings.Unbounded.To_String (Include.Path));
						procedure Process (Target_Page : in Markup.Page) is
						begin
							Process_Mailto (Target_Path, Target_Page);
						end Process;
					begin
						if not Ada.Directories.Exists (Target_Path) then
							Query (Target_Path, Process'Access);
						end if;
					end;
				else
					Markup.Iterate (Item, Process_Element'Access);
				end if;
			end Process_Element;
		begin
			Markup.Iterate (Page.Items, Process_Element'Access);
			if Item_Exists then
				Output ("}" & Ascii.LF);
				Output ("</script>" & Ascii.LF);
			end if;
		end Process_Mailto;
		Navi_Exists : Boolean := False;
		procedure Process_Navigate (Link_Path : in String) is
			Relative_Path : constant String :=
				Pages.Relative_Name (Parent_Path, Link_Path);
			Containing_Directory : constant String :=
				Pages.Containing_Directory (Relative_Path);
		begin
			if not Navi_Exists then
				Navi_Exists := True;
				Output ("<nav>" & Ascii.LF);
			end if;
			Output ("<a href=""");
			if Ada.Directories.Simple_Name (Link_Path) = Pages.Index then
				if Containing_Directory = "" then
					Output ("./");
				else
					Output_In_Attribute (Containing_Directory);
					Output ("/");
				end if;
			else
				Output_In_Attribute (Relative_Path);
				Output (".html");
			end if;
			Output (""">");
			if Containing_Directory = "" then
				Output ("./");
			end if;
			Output_In_HTML (Relative_Path);
			Output (".html");
			Output ("</a>" & Ascii.LF);
		end Process_Navigate;
		Level : Positive := 2;
		Number : array (2 .. 6) of Natural;
		function Label return String is
			Result : Ada.Strings.Unbounded.Unbounded_String;
		begin
			for I in 2 .. Level loop
				if I > 2 then
					Ada.Strings.Unbounded.Append (Result, '-');
				end if;
				Ada.Strings.Unbounded.Append (Result, To_String (Number (I)));
			end loop;
			return Ada.Strings.Unbounded.To_String (Result);
		end Label;
		procedure Process_Index (
			Path : in String;
			Page : in Markup.Page)
		is
			Item_Exists : Boolean := False;
			procedure Process_Block (Item : in Markup.Block'Class) is
			begin
				if Item in Markup.Section.Block then
					declare
						Section : Markup.Section.Block renames Markup.Section.Block (Item);
					begin
						if not Item_Exists then
							if Level = 2 then
								Output ("<nav>");
							end if;
							Output (Ascii.LF & "<ul>" & Ascii.LF);
						end if;
						Number (Level) := Number (Level) + 1;
						Output ("<li><a href=""#");
						if Section.Label = "" then
							Output_In_Attribute (Label);
						else
							Output_In_Attribute (Ada.Strings.Unbounded.To_String (Section.Label));
						end if;
						Output (""">");
						Output_In_HTML (Ada.Strings.Unbounded.To_String (Section.Title));
						Output ("</a>");
						Item_Exists := False;
						Level := Level + 1;
						if Level in Number'Range then
							Number (Level) := 0;
							Markup.Iterate (
								Section.Items,
								Process_Block'Access);
						end if;
						Level := Level - 1;
						if Item_Exists then
							Output ("</ul>" & Ascii.LF);
						end if;
						Output ("</li>" & Ascii.LF);
						Item_Exists := True;
					end;
				elsif Item in Markup.Include.Block then
					declare
						Include : Markup.Include.Block renames Markup.Include.Block (Item);
						Target_Path : constant String :=
							Pages.Compose (
								Pages.Containing_Directory (Path),
								Ada.Strings.Unbounded.To_String (Include.Path));
						procedure Process (Target_Page : in Markup.Page) is
						begin
							if not Item_Exists then
								if Level = 2 then
									Output ("<nav>");
								end if;
								Output (Ascii.LF & "<ul>" & Ascii.LF);
							end if;
							Number (Level) := Number (Level) + 1;
							Output ("<li><a href=""#");
							Output_In_Attribute (Label);
							Output (""">");
							Output_In_HTML (Target_Page.Slice (Target_Page.Title));
							Output ("</a>");
							Item_Exists := False;
							Level := Level + 1;
							if Level in Number'Range then
								Number (Level) := 0;
								Process_Index (Target_Path, Target_Page);
							end if;
							Level := Level - 1;
							if Item_Exists then
								Output ("</ul>" & Ascii.LF);
							end if;
							Output ("</li>" & Ascii.LF);
							Item_Exists := True;
						end Process;
					begin
						if not Ada.Directories.Exists (Target_Path) then
							Query (Target_Path, Process'Access);
						end if;
					end;
				end if;
			end Process_Block;
		begin
			Markup.Iterate (Page.Items, Process_Block'Access);
			if Item_Exists then
				Output ("</ul>" & Ascii.LF);
				Output ("</nav>" & Ascii.LF);
			end if;
		end Process_Index;
		Remark_Index : Natural;
		procedure Process_Article (
			Path : in String;
			Page : in Markup.Page)
		is
			procedure Process_Text (Item : in Markup.Range_Type) is
			begin
				Output_In_HTML (Page.Slice (Item));
			end Process_Text;
			procedure Process_Inline (Item : in Markup.Inline'Class) is
			begin
				if Item in Markup.Abbr.Inline then
					declare
						Abbr : Markup.Abbr.Inline renames Markup.Abbr.Inline (Item);
					begin
						Output ("<abbr title=""");
						Output_In_Attribute (Ada.Strings.Unbounded.To_String (Abbr.Formal));
						Output (""">");
						Markup.Iterate (
							Abbr.Contents,
							Abbr.Items,
							Process_Text'Access,
							Process_Inline'Access);
						Output ("</abbr>");
					end;
				elsif Item in Markup.Anchor.Inline then
					declare
						Anchor : Markup.Anchor.Inline renames Markup.Anchor.Inline (Item);
						Target_Path : constant String :=
							Pages.Reference_Path (
								Pages.Compose (
									Pages.Containing_Directory (Path),
									Ada.Strings.Unbounded.To_String (Anchor.Path)));
						procedure Process (Target_Page : in Markup.Page) is
						begin
							Output_In_HTML (Target_Page.Slice (Target_Page.Title));
						end Process;
						Directory_Exists : Boolean :=
							Target_Path (Target_Path'Last) = '/'
							or else (
								Ada.Directories.Exists (Target_Path)
								and then not Ada.Directories.Exists (
									Ada.Directories.Compose (
										Name => Target_Path,
										Extension => Pages.Source_Extension)));
					begin
						Output ("<a href=""");
						if Directory_Exists then
							Output_In_Attribute (Ada.Strings.Unbounded.To_String (Anchor.Path));
						elsif Ada.Directories.Simple_Name (Target_Path) = Pages.Index then
							-- link to index.html
							Output_In_Attribute (
								Pages.Containing_Directory (
									Pages.Reference_Path (
										Ada.Strings.Unbounded.To_String (Anchor.Path))) & "/");
						else
							Output_In_Attribute (Ada.Directories.Compose (
								Name => Ada.Strings.Unbounded.To_String (Anchor.Path),
								Extension => "html"));
						end if;
						Output (""">");
						if Anchor.Is_Empty then
							if Directory_Exists then
								Output_In_HTML (Ada.Strings.Unbounded.To_String (Anchor.Path));
							else
								Query (Target_Path, Process'Access);
							end if;
						else
							Markup.Iterate (
								Anchor.Contents,
								Anchor.Items,
								Process_Text'Access,
								Process_Inline'Access);
						end if;
						Output ("</a>");
					end;
				elsif Item in Markup.Code.Inline then
					declare
						Code : Markup.Code.Inline renames Markup.Code.Inline (Item);
					begin
						Output ("<code class=""code");
						if Code.Lang /= "" then
							Output (" lang-");
							Output_In_Attribute (Ada.Strings.Unbounded.To_String (Code.Lang));
						end if;
						Output (""">");
						Output_In_HTML (Page.Slice (Item.Contents));
						Output ("</code>");
					end;
				elsif Item in Markup.Date.Inline then
					declare
						Date : Markup.Date.Inline renames Markup.Date.Inline (Item);
						Target_Path : constant String :=
							Pages.Compose (
								Pages.Containing_Directory (Path),
								Ada.Strings.Unbounded.To_String (Date.Path));
						Time : Ada.Calendar.Time := Ada.Directories.Modification_Time (Target_Path);
						Time_Zone : Ada.Calendar.Time_Zones.Time_Offset :=
							Ada.Calendar.Time_Zones.UTC_Time_Offset;
						Image : constant String :=
							Ada.Calendar.Formatting.Image (
								Time,
								Include_Time_Fraction => False,
								Time_Zone => Time_Zone);
						Valid_Image : String := "0000-00-00T00:00+00:00";
					begin
						Valid_Image (1 .. 10) := Image (Image'First .. Image'First + 9);
						Valid_Image (12 .. 16) := Image (Image'First + 11 .. Image'First + 15);
						if Time_Zone < 0 then
							Valid_Image (17) := '-';
						end if;
						Time_Zone := abs Time_Zone;
						Valid_Image (18) := Character'Val (Time_Zone / 600 + Character'Pos ('0'));
						Valid_Image (19) :=
							Character'Val (Time_Zone / 60 rem 10 + Character'Pos ('0'));
						Valid_Image (21) := Character'Val (Time_Zone / 10 rem 6 + Character'Pos ('0'));
						Valid_Image (22) := Character'Val (Time_Zone rem 10 + Character'Pos ('0'));
						Output ("<time datetime=""");
						Output (Valid_Image);
						Output (""">");
						Output (Image (Image'First .. Image'First + 9));
						Output ("</time>");
					end;
				elsif Item in Markup.Delete.Inline then
					declare
						Delete : Markup.Delete.Inline renames Markup.Delete.Inline (Item);
					begin
						Output ("<del>");
						Markup.Iterate (
							Delete.Contents,
							Delete.Items,
							Process_Text'Access,
							Process_Inline'Access);
						Output ("</del>");
					end;
				elsif Item in Markup.Emphasis.Inline then
					declare
						Emphasis : Markup.Emphasis.Inline renames Markup.Emphasis.Inline (Item);
					begin
						for I in Emphasis.Style'Range loop
							for J in 1 .. Emphasis.Style (I) loop
								Output ("<");
								Output (Emphasis_Tag (I));
								Output (">");
							end loop;
						end loop;
						Markup.Iterate (
							Emphasis.Contents,
							Emphasis.Items,
							Process_Text'Access,
							Process_Inline'Access);
						for I in reverse Emphasis.Style'Range loop
							for J in reverse 1 .. Emphasis.Style (I) loop
								Output ("</");
								Output (Emphasis_Tag (I));
								Output (">");
							end loop;
						end loop;
					end;
				elsif Item in Markup.Http.Inline then
					declare
						Http : Markup.Http.Inline renames Markup.Http.Inline (Item);
					begin
						Output ("<a href=""");
						case Http.Scheme is
							when Markup.Http.HTTP => Output_In_Attribute ("http:");
							when Markup.Http.HTTPS => Output_In_Attribute ("https:");
						end case;
						Output_In_Attribute (Ada.Strings.Unbounded.To_String (Http.Link));
						Output (""">");
						if Http.Is_Empty then
							case Http.Scheme is
								when Markup.Http.HTTP => Output_In_HTML ("http:");
								when Markup.Http.HTTPS => Output_In_HTML ("https:");
							end case;
							Output_In_HTML (Ada.Strings.Unbounded.To_String (Http.Link));
						else
							Markup.Iterate (
								Http.Contents,
								Http.Items,
								Process_Text'Access,
								Process_Inline'Access);
						end if;
						Output ("</a>");
					end;
				elsif Item in Markup.Include.Inline then
					declare
						Include : Markup.Include.Inline renames Markup.Include.Inline (Item);
					begin
						Output ("<img src=""");
						Output_In_Attribute (Ada.Strings.Unbounded.To_String (Include.Path));
						Output (""" alt=""");
						Output_In_Attribute (Page.Slice (Include.Contents));
						Output (""" />");
					end;
				elsif Item in Markup.Input.Inline then
					declare
						Input : Markup.Input.Inline renames Markup.Input.Inline (Item);
					begin
						Output ("<kbd class=""input"">");
						Output_In_HTML (Page.Slice (Input.Contents));
						Output ("</kbd>");
					end;
				elsif Item in Markup.Mailto.Inline then
					declare
						Mailto : Markup.Mailto.Inline renames Markup.Mailto.Inline (Item);
					begin
						Mailto_Index := Mailto_Index + 1;
						Output ("<a id=""m");
						Output (To_String (Mailto_Index));
						Output (""">");
						Markup.Iterate (
							Mailto.Contents,
							Mailto.Items,
							Process_Text'Access,
							Process_Inline'Access);
						Output ("</a>");
					end;
				elsif Item in Markup.Output.Inline then
					declare
						The_Output : Markup.Output.Inline renames Markup.Output.Inline (Item);
					begin
						Output ("<samp class=""output"">");
						Output_In_HTML (Page.Slice (The_Output.Contents));
						Output ("</samp>");
					end;
				elsif Item in Markup.Plain.Inline then
					declare
						Plain : Markup.Plain.Inline renames Markup.Plain.Inline (Item);
					begin
						case Plain.Style is
							when Markup.Plain.Proportional =>
								Output ("<span class=""plain"">");
								Output_In_HTML (Page.Slice (Plain.Contents));
								Output ("</span>");
							when Markup.Plain.Fixed =>
								Output ("<span class=""plain fixed"">");
								Output_In_HTML (Page.Slice (Plain.Contents));
								Output ("</span>");
						end case;
					end;
				elsif Item in Markup.Quote.Inline then
					declare
						Quote : Markup.Quote.Inline renames Markup.Quote.Inline (Item);
					begin
						Output ("<q");
						if Quote.Link /= "" then
							Output (" cite=""");
							Output_In_Attribute (Ada.Strings.Unbounded.To_String (Quote.Link));
							Output ("""");
						end if;
						Output (">");
						Markup.Iterate (
							Quote.Contents,
							Quote.Items,
							Process_Text'Access,
							Process_Inline'Access);
						Output ("</q>");
					end;
				elsif Item in Markup.Remark.Inline then
					declare
						Remark : Markup.Remark.Inline renames Markup.Remark.Inline (Item);
					begin
						Remark_Index := Remark_Index + 1;
						if not Remark.Is_Empty then
							Output ("<a href=""#r");
							Output (To_String (Remark_Index));
							Output (""">");
							Markup.Iterate (
								Remark.Contents,
								Remark.Items,
								Process_Text'Access,
								Process_Inline'Access);
							Output ("</a>");
						end if;
						Output ("<a id=""t");
						Output (To_String (Remark_Index));
						Output (""" href=""#r");
						Output (To_String (Remark_Index));
						Output (""" class=""remark"">*");
						Output (To_String (Remark_Index));
						Output ("</a>");
					end;
				elsif Item in Markup.Subject.Inline then
					declare
						Subject : Markup.Subject.Inline renames Markup.Subject.Inline (Item);
					begin
						Output ("<span class=""subject ");
						case Subject.Style is
							when Markup.Subject.Left =>
								Output ("left");
							when Markup.Subject.Right =>
								Output ("right");
						end case;
						Output (""">");
						Markup.Iterate (
							Subject.Contents,
							Subject.Items,
							Process_Text'Access,
							Process_Inline'Access);
						Output ("</span>");
					end;
				else
					raise Program_Error;
				end if;
			end Process_Inline;
			procedure Process_Block (Item : in Markup.Block'Class) is
			begin
				if Item in Markup.Author.Block then
					declare
						Author : Markup.Author.Block renames Markup.Author.Block (Item);
					begin
						Output ("<address>" & Ascii.LF);
						if Author.Link /= "" then
							Output ("<a href=""");
							Output_In_Attribute (Ada.Strings.Unbounded.To_String (Author.Link));
							Output (""">");
						end if;
						Markup.Iterate (
							Author.Contents,
							Author.Items,
							Process_Text'Access,
							Process_Inline'Access);
						if Author.Link /= "" then
							Output ("</a>");
						end if;
						Output (Ascii.LF & "</address>" & Ascii.LF);
					end;
				elsif Item in Markup.Code.Block then
					declare
						Code : Markup.Code.Block renames Markup.Code.Block (Item);
					begin
						Output ("<pre class=""code");
						if Code.Lang /= "" then
							Output (" lang-");
							Output_In_Attribute (Ada.Strings.Unbounded.To_String (Code.Lang));
						end if;
						Output (""">" & Ascii.LF);
						Output ("<code class=""code");
						if Code.Lang /= "" then
							Output (" lang-");
							Output_In_Attribute (Ada.Strings.Unbounded.To_String (Code.Lang));
						end if;
						Output (""">");
						Output_In_HTML (Page.Slice (Item.Contents), Pre => True);
						Output ("</code>" & Ascii.LF & "</pre>" & Ascii.LF);
					end;
				elsif Item in Markup.Delete.Block then
					declare
						Delete : Markup.Delete.Block renames Markup.Delete.Block (Item);
					begin
						Output ("<del>" & Ascii.LF);
						Markup.Iterate (
							Delete.Items,
							Process_Block'Access);
						Output ("</del>" & Ascii.LF);
					end;
				elsif Item in Markup.Emphasis.Block then
					declare
						Emphasis : Markup.Emphasis.Block renames Markup.Emphasis.Block (Item);
					begin
						Output ("<p>" & Ascii.LF);
						for I in Emphasis.Style'Range loop
							for J in 1 .. Emphasis.Style (I) loop
								Output ("<");
								Output (Emphasis_Tag (I));
								Output (">");
							end loop;
						end loop;
						Markup.Iterate (
							Emphasis.Contents,
							Emphasis.Items,
							Process_Text'Access,
							Process_Inline'Access);
						for I in reverse Emphasis.Style'Range loop
							for J in reverse 1 .. Emphasis.Style (I) loop
								Output ("</");
								Output (Emphasis_Tag (I));
								Output (">");
							end loop;
						end loop;
						Output (Ascii.LF & "</p>" & Ascii.LF);
					end;
				elsif Item in Markup.History.Block then
					null;
				elsif Item in Markup.Include.Block then
					declare
						Include : Markup.Include.Block renames Markup.Include.Block (Item);
						Target_Path : constant String :=
							Pages.Compose (
								Pages.Containing_Directory (Path),
								Ada.Strings.Unbounded.To_String (Include.Path));
						procedure Process (Target_Page : in Markup.Page) is
						begin
							Number (Level) := Number (Level) + 1;
							Output ("<section>" & Ascii.LF);
							Output ("<h" & To_String (Level) & " id=""");
							Output_In_Attribute (Label);
							Output (""">");
							Output_In_HTML (Target_Page.Slice (Target_Page.Title));
							Output ("</h" & To_String (Level) & ">" & Ascii.LF);
							Level := Level + 1;
							if Level in Number'Range then
								Number (Level) := 0;
							end if;
							Process_Article (Target_Path, Target_Page);
							Level := Level - 1;
							Output ("</section>" & Ascii.LF);
						end Process;
					begin
						if Ada.Directories.Exists (Target_Path) then
							Output ("<p>" & Ascii.LF);
							Output ("<img src=""");
							Output_In_Attribute (Ada.Strings.Unbounded.To_String (Include.Path));
							Output (""" alt=""");
							Output_In_Attribute (Page.Slice (Include.Contents));
							Output (""" />" & Ascii.LF);
							Output ("</p>" & Ascii.LF);
						else
							Query (Target_Path, Process'Access);
						end if;
					end;
				elsif Item in Markup.Input.Block then
					declare
						Input : Markup.Input.Block renames Markup.Input.Block (Item);
					begin
						Output ("<pre class=""input"">" & Ascii.LF);
						Output ("<kbd class=""input"">");
						Output_In_HTML (Page.Slice (Input.Contents), Pre => True);
						Output ("</kbd>" & Ascii.LF);
						Output ("</pre>" & Ascii.LF);
					end;
				elsif Item in Markup.List.Block then
					declare
						List : Markup.List.Block renames Markup.List.Block (Item);
						procedure Close is
						begin
							case List.Style is
								when Markup.List.None | Markup.List.Unordered =>
									Output ("</ul>" & Ascii.LF);
								when Markup.List.Ordered =>
									Output ("</ol>" & Ascii.LF);
							end case;
						end Close;
						Level : Integer := -1;
						I : Markup.List.Lists.Cursor := List.Items.First;
					begin
						while Markup.List.Lists.Has_Element (I) loop
							declare
								Item : Markup.List.Item
									renames List.Items.Constant_Reference (I);
							begin
								if Item.Level > Level then
									if Level >= 0 then
										Output ((1 => Ascii.LF));
									end if;
									case List.Style is
										when Markup.List.None =>
											Output ("<ul class=""none"">" & Ascii.LF);
										when Markup.List.Unordered =>
											Output ("<ul>" & Ascii.LF);
										when Markup.List.Ordered =>
											Output ("<ol>" & Ascii.LF);
									end case;
									Level := Level + 1;
								else
									while Item.Level < Level loop
										Output ("</li>" & Ascii.LF);
										Close;
										Level := Level - 1;
									end loop;
									Output ("</li>" & Ascii.LF);
								end if;
								Output ("<li>");
								Markup.Iterate (
									Item.Contents,
									Item.Items,
									Process_Text'Access,
									Process_Inline'Access);
							end;
							Markup.List.Lists.Next (I);
						end loop;
						for I in 0 .. Level loop
							Output ("</li>" & Ascii.LF);
							Close;
						end loop;
					end;
				elsif Item in Markup.Output.Block then
					declare
						The_Output : Markup.Output.Block renames Markup.Output.Block (Item);
					begin
						Output ("<pre class=""output"">" & Ascii.LF);
						Output ("<samp class=""output"">");
						Output_In_HTML (Page.Slice (The_Output.Contents), Pre => True);
						Output ("</samp>" & Ascii.LF);
						Output ("</pre>" & Ascii.LF);
					end;
				elsif Item in Markup.Paragraph.Block then
					declare
						Paragraph : Markup.Paragraph.Block renames Markup.Paragraph.Block (Item);
					begin
						Output ("<p>" & Ascii.LF);
						Markup.Iterate (
							Paragraph.Contents,
							Paragraph.Items,
							Process_Text'Access,
							Process_Inline'Access);
						Output (Ascii.LF & "</p>" & Ascii.LF);
					end;
				elsif Item in Markup.Plain.Block then
					declare
						Plain : Markup.Plain.Block renames Markup.Plain.Block (Item);
					begin
						case Plain.Style is
							when Markup.Plain.Proportional =>
								Output ("<p class=""plain"">" & Ascii.LF);
								Output_In_HTML (Page.Slice (Plain.Contents));
								Output (Ascii.LF & "</p>" & Ascii.LF);
							when Markup.Plain.Fixed =>
								Output ("<pre class=""plain fixed"">" & Ascii.LF);
								Output_In_HTML (Page.Slice (Plain.Contents), Pre => True);
								Output (Ascii.LF & "</pre>" & Ascii.LF);
						end case;
					end;
				elsif Item in Markup.Quote.Block then
					declare
						Quote : Markup.Quote.Block renames Markup.Quote.Block (Item);
					begin
						Output ("<blockquote");
						if Quote.Link /= "" then
							Output (" cite=""");
							Output_In_Attribute (Ada.Strings.Unbounded.To_String (Quote.Link));
							Output ("""");
						end if;
						Output (">" & Ascii.LF);
						Markup.Iterate (
							Quote.Items,
							Process_Block'Access);
						Output ("</blockquote>" & Ascii.LF);
					end;
				elsif Item in Markup.Section.Block then
					declare
						Section : Markup.Section.Block renames Markup.Section.Block (Item);
					begin
						Output ("<section>" & Ascii.LF);
						if Level in Number'Range then
							Number (Level) := Number (Level) + 1;
							Output ("<h" & To_String (Level) & " id=""");
							if Section.Label = "" then
								Output_In_Attribute (Label);
							else
								Output_In_Attribute (Ada.Strings.Unbounded.To_String (Section.Label));
							end if;
							Output (""">");
							Output_In_HTML (Ada.Strings.Unbounded.To_String (Section.Title));
							Output ("</h" & To_String (Level) & ">" & Ascii.LF);
						else
							Output ("<p class=""h" & To_String (Level) & """>");
							Output_In_HTML (Ada.Strings.Unbounded.To_String (Section.Title));
							Output ("</p>" & Ascii.LF);
						end if;
						Level := Level + 1;
						if Level in Number'Range then
							Number (Level) := 0;
						end if;
						Markup.Iterate (
							Section.Items,
							Process_Block'Access);
						Level := Level - 1;
						Output ("</section>" & Ascii.LF);
					end;
				elsif Item in Markup.Table.Block then
					declare
						Table : Markup.Table.Block renames Markup.Table.Block (Item);
						Has_Line : Boolean := False;
					begin
						Output ("<table>" & Ascii.LF);
						Output ("<tbody>" & Ascii.LF);
						declare
							I : Markup.Table.Lists.Cursor := Table.Items.First;
						begin
							while Markup.Table.Lists.Has_Element (I) loop
								declare
									Item : Markup.Table.Item
										renames Table.Items.Constant_Reference (I);
								begin
									if Item.First then
										if Has_Line then
											Output ("</tr>" & Ascii.LF);
										end if;
										Has_Line := True;
										Output ("<tr>");
									end if;
									Output ("<td>");
									Markup.Iterate (
										Item.Contents,
										Item.Items,
										Process_Text'Access,
										Process_Inline'Access);
									Output ("</td>");
								end;
								I := Markup.Table.Lists.Next (I);
							end loop;
						end;
						if Has_Line then
							Output ("</tr>" & Ascii.LF);
						end if;
						Output ("</tbody>" & Ascii.LF);
						Output ("</table>" & Ascii.LF);
					end;
				else
					raise Program_Error;
				end if;
			end Process_Block;
		begin
			Markup.Iterate (Page.Items, Process_Block'Access);
		end Process_Article;
		Aside_Exists : Boolean := False;
		procedure Process_History (
			Path : in String;
			Page : in Markup.Page)
		is
			Item_Exists : Boolean := False;
			procedure Process_Element (Item : in Markup.Element'Class) is
			begin
				if Item in Markup.History.Block then
					declare
						History : Markup.History.Block renames Markup.History.Block (Item);
					begin
						if not Aside_Exists then
							Output ("<aside>" & Ascii.LF);
							Aside_Exists := True;
						end if;
						if not Item_Exists then
							Output ("<p class=""history"">" & Ascii.LF);
						else
							Output ("<br />" & Ascii.LF);
						end if;
						Output_In_HTML (Page.Slice (History.Contents));
						Item_Exists := True;
					end;
				elsif Item in Markup.Include.Block then
					declare
						Include : Markup.Include.Block renames Markup.Include.Block (Item);
						Target_Path : constant String :=
							Pages.Compose (
								Pages.Containing_Directory (Path),
								Ada.Strings.Unbounded.To_String (Include.Path));
						procedure Process (Target_Page : in Markup.Page) is
						begin
							Process_History (Target_Path, Target_Page);
						end Process;
					begin
						if not Ada.Directories.Exists (Target_Path) then
							Query (Target_Path, Process'Access);
						end if;
					end;
				else
					Markup.Iterate (Item, Process_Element'Access);
				end if;
			end Process_Element;
		begin
			Markup.Iterate (Page.Items, Process_Element'Access);
			if Item_Exists then
				Output ("</p>" & Ascii.LF);
			end if;
		end Process_History;
		procedure Process_Remark (
			Path : in String;
			Page : in Markup.Page)
		is
			Item_Exists : Boolean := False;
			procedure Process_Element (Item : in Markup.Element'Class) is
			begin
				if Item in Markup.Remark.Inline then
					declare
						Remark : Markup.Remark.Inline renames Markup.Remark.Inline (Item);
					begin
						if not Aside_Exists then
							Output ("<aside>" & Ascii.LF);
							Aside_Exists := True;
						end if;
						if not Item_Exists then
							Output ("<p class=""remark"">" & Ascii.LF);
						else
							Output ("<br />" & Ascii.LF);
						end if;
						Remark_Index := Remark_Index + 1;
						Output ("<a id=""r");
						Output (To_String (Remark_Index));
						Output (""" href=""#t");
						Output (To_String (Remark_Index));
						Output (""" class=""remark"">*");
						Output (To_String (Remark_Index));
						Output ("</a> ");
						Output_In_HTML (Ada.Strings.Unbounded.To_String (Remark.Note));
						Item_Exists := True;
					end;
				elsif Item in Markup.Include.Block then
					declare
						Include : Markup.Include.Block renames Markup.Include.Block (Item);
						Target_Path : constant String :=
							Pages.Compose (
								Pages.Containing_Directory (Path),
								Ada.Strings.Unbounded.To_String (Include.Path));
						procedure Process (Target_Page : in Markup.Page) is
						begin
							Process_Remark (Target_Path, Target_Page);
						end Process;
					begin
						if not Ada.Directories.Exists (Target_Path) then
							Query (Target_Path, Process'Access);
						end if;
					end;
				else
					Markup.Iterate (Item, Process_Element'Access);
				end if;
			end Process_Element;
		begin
			Markup.Iterate (Page.Items, Process_Element'Access);
			if Item_Exists then
				Output ("</p>" & Ascii.LF);
			end if;
		end Process_Remark;
		Title : constant String := Page.Slice (Page.Title);
	begin
		Output ("<?xml version=""1.0"" encoding=""UTF-8""?>" & Ascii.LF);
		Output ("<!DOCTYPE html>" & Ascii.LF);
		Output (
			"<html xmlns=""http://www.w3.org/1999/xhtml"" lang=""en"" xml:lang=""en"">"
			& Ascii.LF);
		Output ("<head>" & Ascii.LF);
		Output ("<meta name=""generator"" content=""nightscream"" />" & Ascii.LF);
		if Style_Sheet /= "" then
			Output ("<link rel=""stylesheet"" type=""text/css"" href=""");
			Output_In_Attribute (Style_Sheet);
			Output (""" />" & Ascii.LF);
		end if;
		if Title /= "" then
			Output ("<title>");
			Output_In_HTML (Title);
			Output ("</title>" & Ascii.LF);
		end if;
		Mailto_Index := 0;
		Process_Mailto (Path, Page);
		Output ("</head>" & Ascii.LF);
		Output ("<body");
		if Mailto_Index > 0 then
			Output (" onload=""set_m();""");
		end if;
		Output (">" & Ascii.LF);
		Output ("<header>" & Ascii.LF);
		Navigate (Process_Navigate'Access);
		if Navi_Exists then
			Output ("</nav>" & Ascii.LF);
		end if;
		if Title /= "" then
			Output ("<h1>");
			Output_In_HTML (Title);
			Output ("</h1>" & Ascii.LF);
		end if;
		Output ("</header>" & Ascii.LF);
		Output ("<article>" & Ascii.LF);
		Number (Level) := 0;
		Process_Index (Path, Page);
		Number (Level) := 0;
		Remark_Index := 0;
		Mailto_Index := 0;
		Process_Article (Path, Page);
		Output ("</article>" & Ascii.LF);
		Output ("<footer>" & Ascii.LF);
		Process_History (Path, Page);
		Remark_Index := 0;
		Process_Remark (Path, Page);
		if Aside_Exists then
			Output ("</aside>" & Ascii.LF);
		end if;
		Output ("</footer>" & Ascii.LF);
		Output ("</body>" & Ascii.LF);
		Output ("</html>" & Ascii.LF);
	end Output;
	
end Scream.Output.HTML;
