with Ada.Calendar.Formatting;
with Ada.Directories;
with Ada.IO_Exceptions;
with Ada.Streams.Stream_IO;
with Ada.Strings.Fixed;
with zlib.Streams;
package body Scream.Pages is
	use type Ada.Calendar.Time;
	
	function Exclude (S : String; Filter : String) return String is
		Result : String (S'Range);
		Last : Natural := S'First - 1;
	begin
		for I in S'Range loop
			for J in Filter'Range loop
				if Filter (J) = S (I) then
					goto Found;
				end if;
			end loop;
			Last := Last + 1;
			Result (Last) := S (I);
			<<Found>> null;
		end loop;
		return Result (S'First .. Last);
	end Exclude;
	
	Signature : constant String := "TSPI";
	
	function Get_Page_Info (Path : String) return Page_Info is
		Storage_Dir : constant String :=
			Ada.Directories.Compose (Containing_Directory (Path), Storage_Directory);
		Base_Name : constant String := Ada.Directories.Base_Name (Path);
		Info_Name : constant String :=
			Ada.Directories.Compose (
				Containing_Directory => Storage_Dir,
				Name => Base_Name,
				Extension => Info_Extension);
		File : Ada.Streams.Stream_IO.File_Type;
	begin
		return Result : Page_Info do
			begin
				Ada.Streams.Stream_IO.Open (
					File,
					Ada.Streams.Stream_IO.In_File,
					Name => Info_Name);
				declare
					Stream : Ada.Streams.Stream_IO.Stream_Access :=
						Ada.Streams.Stream_IO.Stream (File);
					Signature_Read : String (Signature'Range);
				begin
					String'Read (Stream, Signature_Read);
					if Signature_Read /= Signature then
						raise Ada.IO_Exceptions.Data_Error;
					end if;
					Ada.Calendar.Time'Read (Stream, Result.Creation_Time);
					Ada.Calendar.Time'Read (Stream, Result.Build_Time);
					Page_Sets.Set'Read (Stream, Result.Reference);
				exception
					when others =>
						Ada.Streams.Stream_IO.Close (File);
						raise;
				end;
				Ada.Streams.Stream_IO.Close (File);
				Result.Is_New := False;
			exception
				when Ada.IO_Exceptions.Name_Error =>
					declare
						Source : constant String :=
							Ada.Directories.Compose (Name => Path, Extension => Source_Extension);
						Modification_Time : Ada.Calendar.Time :=
							Ada.Directories.Modification_Time (Source);
					begin
						Result.Creation_Time := Modification_Time;
						Result.Build_Time := Modification_Time - 1.0;
						Result.Reference := Page_Sets.Empty_Set;
						Result.Is_New := True;
					end;
			end;
		end return;
	end Get_Page_Info;
	
	procedure Save_Page_Info (
		Path : String;
		Info : Page_Info;
		Source : String;
		Modified : Boolean)
	is
		Storage_Dir : constant String :=
			Ada.Directories.Compose (Containing_Directory (Path), Storage_Directory);
		Base_Name : constant String := Ada.Directories.Base_Name (Path);
		Info_Name : constant String :=
			Ada.Directories.Compose (
				Containing_Directory => Storage_Dir,
				Name => Base_Name,
				Extension => Info_Extension);
		File : Ada.Streams.Stream_IO.File_Type;
	begin
		if not Ada.Directories.Exists (Storage_Dir) then
			Ada.Directories.Create_Directory (Storage_Dir);
		end if;
		-- page info
		Ada.Streams.Stream_IO.Create (File, Name => Info_Name);
		declare
			Stream : Ada.Streams.Stream_IO.Stream_Access :=
				Ada.Streams.Stream_IO.Stream (File);
		begin
			String'Write (Stream, Signature);
			Ada.Calendar.Time'Write (Stream, Info.Creation_Time);
			Ada.Calendar.Time'Write (Stream, Info.Build_Time);
			Page_Sets.Set'Write (Stream, Info.Reference);
		end;
		Ada.Streams.Stream_IO.Close (File);
		-- compress source
		if Modified then
			declare
				Build_Time_Image : String :=
					Exclude (
						Ada.Calendar.Formatting.Image (Info.Build_Time),
						(':', '-', ' '));
				Backup_Name : constant String :=
					Ada.Directories.Compose (
						Containing_Directory => Storage_Dir,
						Name => Build_Time_Image & " " & Base_Name,
						Extension => Backup_Extension);
			begin
				Ada.Streams.Stream_IO.Create (File, Name => Backup_Name);
				declare
					Z_Stream : aliased zlib.Streams.Out_Type :=
						zlib.Streams.Open (
							ZLib.Streams.Stream_Access (Ada.Streams.Stream_IO.Stream (File)),
							Header => ZLib.GZip);
				begin
					String'Write (zlib.Streams.Stream (Z_Stream), Source);
					ZLib.Streams.Finish (Z_Stream);
				end;
				Ada.Streams.Stream_IO.Close (File);
			end;
		end if;
	end Save_Page_Info;
	
	function Reference_Path (Path : String) return String is
	begin
		if Path (Path'Last) = '/'
			and then Ada.Directories.Exists (
				Ada.Directories.Compose (Path, Index, Source_Extension))
		then
			return Ada.Directories.Compose (Path, Index);
		else
			return Path;
		end if;
	end Reference_Path;
	
	function Containing_Directory (Name : String) return String is
	begin
		return Ada.Directories.Containing_Directory (Name);
	exception
		when Ada.Directories.Use_Error => -- Name is simple name
			return "";
	end Containing_Directory;
	
	function Compose (Containing_Directory : String; Name : String)
		return String
	is
		CF : Positive := Containing_Directory'First;
		CL : Natural := Containing_Directory'Last;
		I : Positive := Name'First;
		Saved_CL : Natural;
	begin
		while CF + 1 <= CL
			and then Containing_Directory (CF) = '.'
			and then Containing_Directory (CF + 1) = '/'
		loop
			CF := CF + 2;
		end loop;
		if CF = CL and then Containing_Directory (CF) = '.' then
			CF := CF + 1;
		end if;
		loop
			if I + 1 <= Name'Last and then Name (I) = '.' and then Name (I + 1) = '/' then
				I := I + 2;
			elsif I + 2 <= Name'Last
				and then Name (I) = '.'
				and then Name (I + 1) = '.'
				and then Name (I + 2) = '/'
			then
				Saved_CL := CL;
				for J in reverse CF + 1 .. CL loop
					if Containing_Directory (J) = '/' then
						CL := J - 1;
						I := I + 3;
						exit;
					end if;
				end loop;
				if CL = Saved_CL then
					if CF > CL or else Containing_Directory (CF) = '/' then
						exit;
					else
						CL := CF - 1;
						I := I + 3;
					end if;
				end if;
			else
				exit;
			end if;
		end loop;
		if CF <= CL and then Containing_Directory (CL) /= '/' then
			return Containing_Directory (CF .. CL) & '/' & Name (I .. Name'Last);
		else
			return Containing_Directory (CF .. CL) & Name (I .. Name'Last);
		end if;
	end Compose;
	
	function Relative_Name (
		Base_Directory : String;
		Name : String;
		Drive_Support : Boolean := False)
		return String
	is
		function "*" (Left : Natural; Right : String) return String
			renames Ada.Strings.Fixed."*";
		type Path_Kind is (Relative, Absolute, Drive);
		pragma Discard_Names(Path_Kind);
		function Get_Path_Kind(Name : String) return Path_Kind is
		begin
			if Name'Length = 0 then
				return Relative; -- think as "."
			elsif Name(Name'First) = '/' then
				return Absolute;
			else
				if Drive_Support then
					for I in Name'First .. Name'Last - 1 loop
						if Name(I) = ':' and then Name(I + 1) = '/' then
							return Drive;
						elsif Name(I) = '/' or else Name(I) = '.' then
							exit;
						end if;
					end loop;
				end if;
				return Relative;
			end if;
		end Get_Path_Kind;
		Base_Kind : constant Path_Kind := Get_Path_Kind(Base_Directory);
		Name_Kind : constant Path_Kind := Get_Path_Kind(Name);
		Base_First : Positive := Base_Directory'First;
		Name_First : Positive := Name'First;
		I, Level : Natural;
	begin
		if Base_Kind = Relative then
			if Name_Kind /= Relative then
				return Name;
			end if;
		else
			if Name_Kind /= Base_Kind then
				raise Constraint_Error;
			end if;
			if Name_Kind = Absolute then
				Base_First := Base_First + 1; -- skip first "/"
				Name_First := Name_First + 1;
			end if;
		end if;
		Taking : loop
			if Base_First > Base_Directory'Last then
				return Name(Name_First .. Name'Last);
			else
				I := 0;
				loop
					exit Taking when I + Name_First > Name'Last;
					exit when Name(I + Name_First) = '/';
					I := I + 1;
				end loop;
				if Base_Directory'Last + 1 = I + Base_First
					or else (
						Base_Directory'Last >= I + Base_First
						and then Base_Directory (I + Base_First) = '/'
						and then Base_Directory (Base_First .. I - 1 + Base_First) =
							Name (Name_First .. I - 1 + Name_First))
				then
					Base_First := I + 1 + Base_First;
					Name_First := I + 1 + Name_First;
				else
					exit Taking;
				end if;
			end if;
		end loop Taking;
		if (Name_Kind = Drive and then Name_First = Name'First)
			or else (
				Base_First = Base_Directory'Last
				and then Base_Directory (Base_First) = '.')
		then
			return Name (Name_First .. Name'Last);
		end if;
		Level := 1;
		for I in Base_First .. Base_Directory'Last loop
			if Base_Directory(I) = '/' then
				if I > Base_First
					and then Base_Directory (I - 1) = '.'
					and then (I = Base_First + 1 or else Base_Directory (I - 2) = '/')
				then
					null;
				else
					Level := Level + 1;
				end if;
			end if;
		end loop;
		return Level * "../" & Name(Name_First .. Name'Last);
	end Relative_Name;
	
end Scream.Pages;
